#include "Preconditioner.h"

namespace AMDiS {

  template<typename VectorType>
  ODirSolver<VectorType>::ODirSolver(std::string name)
    : OEMSolver<VectorType>(name), 
      Ad(NULL), d(NULL), d_old(NULL), r(NULL), h(NULL)
  {
  }

  template<typename VectorType>
  ODirSolver<VectorType>::~ODirSolver()
  {}

  template<typename VectorType>
  void ODirSolver<VectorType>::init()
  {
    Ad    = this->vectorCreator->create();
    d     = this->vectorCreator->create();
    d_old = this->vectorCreator->create();
    r     = this->vectorCreator->create();
    h     = this->vectorCreator->create();
  }

  template<typename VectorType>
  void ODirSolver<VectorType>::exit()
  {
    this->vectorCreator->free(Ad);
    this->vectorCreator->free(d);
    this->vectorCreator->free(d_old);
    this->vectorCreator->free(r);
    this->vectorCreator->free(h);
  }

  template<typename VectorType>
  int ODirSolver<VectorType>::solveSystem(MatVecMultiplier<VectorType> *matVec,
					  VectorType *x, VectorType *b, bool reuseMatrix)
  {
    FUNCNAME("ODirSolver::solveSystem");

    double res, old_res = -1.0;
    double dad, dad_old, rd, alpha, gamma, sigma;
    int    iter;
    VectorType *swap = NULL;
    const double TOL = 1.e-30;

    /*--------------------------------------------------------------------------*/
    /*---  Initalization  ------------------------------------------------------*/
    /*--------------------------------------------------------------------------*/

    if (norm(b) < TOL)
      {
	INFO(this->info,2)("b == 0, x = 0 is the solution of the linear system\n");
	set(*x, 0.0);
	return(0);
      }

    // r = Ax
    matVec->matVec(NoTranspose, *x, *r);

    // r = -b + r
    *r -= *b;

    // h = r
    *h = *r;

    if (this->leftPrecon) {
      this->leftPrecon->precon(h);
    }

    // d = h
    *d = *h;

    // d_old = 0;
    set(*d_old, 0.0);

    dad_old = 1.0;

    /*--------------------------------------------------------------------------*/
    /*---  check initial residual  ---------------------------------------------*/
    /*--------------------------------------------------------------------------*/

    res = norm(r);

    START_INFO();
    if (SOLVE_INFO(0, res, &old_res))
      return(0);

    /*--------------------------------------------------------------------------*/
    /*---  Iteration  ----------------------------------------------------------*/
    /*--------------------------------------------------------------------------*/

    for (iter = 1; iter <= this->max_iter; iter++)
      {
	/*--------------------------------------------------------------------------*/
	/*---  calculate r.d, A.d and d.A.d  ---------------------------------------*/
	/*--------------------------------------------------------------------------*/
	// rd = r*d
	rd = *r * *d;

	// Ad = A*d
	matVec->matVec(NoTranspose, *d, *Ad);

	// dad = d*Ad
	dad = *d * *Ad;

	if (abs(dad) < TOL)
	  {
	    BREAK_INFO("(Ad,d)_2 = 0", iter, res, &old_res);
	    return(iter);
	  }

	/*--------------------------------------------------------------------------*/
	/*---   update x and calculate new r  --------------------------------------*/
	/*--------------------------------------------------------------------------*/
	alpha = rd/dad;

	// x = -alpha d + x
	axpy(-alpha, *d, *x);

	// r = -alpha Ad + r
	axpy(-alpha, *Ad, *r);

	res = norm(r);
	if (SOLVE_INFO(iter, res, &old_res))
	  return(iter);

	// h = Ad
	*h = *Ad;

	if (this->leftPrecon) this->leftPrecon->precon(h);

	/*--------------------------------------------------------------------------*/
	/*---  calculate gamma and sigma  ------------------------------------------*/
	/*--------------------------------------------------------------------------*/

	// gamma = h*Ad/dad
	gamma = (*h * *Ad)/dad;
	sigma = dad/dad_old;

	/*--------------------------------------------------------------------------*/
	/*---  compute new d and save old values  ----------------------------------*/
	/*--------------------------------------------------------------------------*/
    
	// swap d and d_old
	//swap(*d, *d_old);
	swap = d;
	d = d_old;
	d_old = swap;

	// d = -sigma * d
	*d *= -sigma;

	// d = -gamma d_old + d
	axpy(-gamma, *d_old, *d);

	// d = h + d
	*d += *h;

	dad_old= dad;
      }

    return 0; // never reached !
  }

}
