// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file ProblemScal.h */

#ifndef AMDIS_PROBLEMSCAL_H
#define AMDIS_PROBLEMSCAL_H

#include <list>

#include "Global.h"
#include "ProblemStatBase.h"
#include "MemoryManager.h"
#include "AbstractFunction.h"
#include "FixVec.h"
#include "Boundary.h"
#include "StandardProblemIteration.h"
#include "ProblemVec.h"

#ifdef HAVE_PARALLEL_AMDIS
#include "petscao.h"
#endif

namespace AMDiS {

  class Operator;
  class DOFMatrix;
  class FiniteElemSpace;
  class Estimator;
  template<typename T> class OEMSolver;
  template<typename T> class Preconditioner;
  template<typename T> class MatVecMultiplier;
  template<typename T> class DOFVector;
  class Marker;
  class AdaptInfo;
  class ElInfo;
  class FileWriterInterface;
  class RefinementManager;
  class CoarseningManager;

  class ProblemScal : public ProblemStatBase,
		      public StandardProblemIteration
  {
  public:
    MEMORY_MANAGED(ProblemScal);

    ProblemScal(const std::string &nameStr, 
		ProblemIterationInterface *problemIteration = NULL)
      : StandardProblemIteration(this),
        name(nameStr),
	feSpace_(NULL),
	mesh(NULL),
	marker(NULL),
	estimator(NULL),
	solver(NULL),
	solution(NULL),
	rhs(NULL),
	systemMatrix(NULL),
	matVec(NULL),
	leftPrecon(NULL),
	rightPrecon(NULL),
	useGetBound(true),
	refinementManager(NULL),
	coarseningManager(NULL),
#ifdef HAVE_PARALLEL_AMDIS
        applicationOrdering(NULL),
        nRankDOFs(0),
#endif
	info_(10)
    {}

    /// Destructor
    virtual ~ProblemScal();

    /// Initialisation of the problem.
    virtual void initialize(Flag initFlag,
			    ProblemScal *adoptProblem = NULL,
			    Flag adoptFlag = INIT_NOTHING);

    /// Used in \ref initialize().
    virtual void createMesh();

    /// Used in \ref initialize().
    virtual void createFESpace();

    /// Used in \ref initialize().
    virtual void createMatricesAndVectors();

    /// Used in \ref initialize().
    virtual void createSolver();

    /// Used in \ref initialize().
    virtual void createEstimator();

    /// Used in \ref initialize().
    virtual void createMarker();

    /// Used in \ref initialize().
    virtual void createFileWriter();

    /** \brief
     * Implementation of ProblemStatBase::solve(). Deligates the solving
     * of problems system to \ref solver.
     */
    virtual void solve(AdaptInfo *adaptInfo, bool fixedMatrix);

    /** \brief
     * Implementation of ProblemStatBase::estimate(). Deligates the estimation
     * to \ref estimator.
     */
    virtual void estimate(AdaptInfo *adaptInfo);

    /** \brief
     * Implementation of ProblemStatBase::markElements().
     * Deligated to \ref adapt.
     */
    virtual Flag markElements(AdaptInfo *adaptInfo);

    /** \brief
     * Implementation of ProblemStatBase::refineMesh(). Deligated to the
     * RefinementManager of \ref adapt.
     */
    virtual Flag refineMesh(AdaptInfo *adaptInfo);

    /** \brief
     * Implementation of ProblemStatBase::coarsenMesh(). Deligated to the
     * CoarseningManager of \ref adapt.
     */
    virtual Flag coarsenMesh(AdaptInfo *adaptInfo);

    /** \brief
     * Implementation of ProblemStatBase::buildBeforeRefine().
     * Does nothing here.
     */
    virtual void buildBeforeRefine(AdaptInfo *adaptInfo, Flag) {}

    /** \brief
     * Implementation of ProblemStatBase::buildBeforeCoarsen().
     * Does nothing here.
     */
    virtual void buildBeforeCoarsen(AdaptInfo *adaptInfo, Flag) {}

    /** \brief
     * Implementation of ProblemStatBase::buildAfterCoarsen().
     * Assembles \ref A and \ref rhs.
     */
    virtual void buildAfterCoarsen(AdaptInfo *adaptInfo, Flag flag,
				   bool assembleMatrix = true,
				   bool assembleVector = true);

    /// Returns number of managed problems
    virtual int getNumProblems() { 
      return 1; 
    }

    /// Implementation of ProblemStatBase::getNumComponents()
    virtual int getNumComponents() { 
      return 1; 
    }

    /** \brief
     * Returns the problem with the given number. If only one problem
     * is managed by this master problem, the number hasn't to be given.
     */
    virtual ProblemStatBase *getProblem(int number = 0) { 
      return this; 
    }

    /// Writes output files.
    void writeFiles(AdaptInfo *adaptInfo, bool force);

    /// Interpolates fct to \ref solution.
    void interpolInitialSolution(AbstractFunction<double, WorldVector<double> > *fct);

    /// Adds an Operator to \ref systemMatrix.
    void addMatrixOperator(Operator *op, 
			   double *factor = NULL, double *estFactor = NULL);

    /// Adds an Operator to \ref rhs.
    void addVectorOperator(Operator *op, 
			   double *factor = NULL, double *estFactor = NULL);

    /// Adds dirichlet boundary conditions.
    virtual void addDirichletBC(BoundaryType type, 
				AbstractFunction<double, WorldVector<double> >* b);

    /// Adds dirichlet boundary conditions.
    virtual void addDirichletBC(BoundaryType type, 
				DOFVector<double> *vec);

    /// Adds robin boundary conditions.
    virtual void addRobinBC(BoundaryType type, 
			    AbstractFunction<double, WorldVector<double> > *n,
			    AbstractFunction<double, WorldVector<double> > *r);

    /// Adds robin boundary conditions.
    virtual void addRobinBC(BoundaryType type, 
			    DOFVector<double> *n,
			    DOFVector<double> *r);

    /// Adds Neumann boundary conditions.
    virtual void addNeumannBC(BoundaryType type, 
			      AbstractFunction<double, WorldVector<double> > *n);


    /// Adds periodic boundary conditions.
    virtual void addPeriodicBC(BoundaryType type);
    
#ifdef HAVE_PARALLEL_AMDIS
    /// Sets the petsc application ordering object to map dof indices.
    void useApplicationOrdering(AO *ao, int nDofs);
#endif

    // ===== getting-methods ======================================================

    /** \name getting methods
     * \{ 
     */

    /// Returns \ref solution.
    inline DOFVector<double>* getSolution() { 
      return solution; 
    }

    /// Returns \ref rhs.
    inline DOFVector<double>* getRHS() { 
      return rhs; 
    }

    /// Returns \ref systemMatrix.
    inline DOFMatrix *getSystemMatrix() { 
      return systemMatrix; 
    }

    /// Returns \ref mesh
    inline Mesh* getMesh() { 
      return mesh; 
    }

    /// Returns \ref feSpace_.
    inline FiniteElemSpace* getFESpace() { 
      return feSpace_; 
    }

    /// Returns \ref estimator_.
    inline Estimator* getEstimator() { 
      return estimator; 
    }

    /// Returns \ref refinementManager.
    inline RefinementManager* getRefinementManager() { 
      return refinementManager; 
    }

    /// Returns \ref solver.
    inline OEMSolver<DOFVector<double> >* getSolver() { 
      return solver; 
    }

    /// Returns \ref marker_.
    inline Marker *getMarker() { 
      return marker; 
    }

    /// Returns \ref name.
    virtual inline const std::string& getName() { 
      return name; 
    }

    /// Returns \ref useGetBound.
    inline bool getBoundUsed() { 
      return useGetBound; 
    }

    /// Returns \ref leftPrecon.
    inline Preconditioner<DOFVector<double> > *getLeftPrecon() { 
      return leftPrecon; 
    }

    /// Returns \ref rightPrecon.
    inline Preconditioner<DOFVector<double> > *getRightPrecon() { 
      return rightPrecon; 
    }

    ///
    inline CoarseningManager *getCoarseningManager() {
      return coarseningManager;
    }

    /** \} */

    // ===== setting-methods ======================================================

    /** \name setting methods
     * \{ 
     */

    /// Sets \ref mesh
    void setMeshFromProblemVec(ProblemVec* pv, int i = 0); 

    /// Sets \ref feSpace_.
    inline void setFESpace(FiniteElemSpace* fe) { 
      feSpace_ = fe; 
    }

    /// Sets \ref estimator_.
    inline void setEstimator(Estimator* est) { 
      estimator = est; 
    }

    /// Sets \ref solver.
    inline void setSolver(OEMSolver<DOFVector<double> >* sol) { 
      solver = sol; 
    }

    /// Sets \ref leftPrecon.
    inline void setLeftPrecon(Preconditioner<DOFVector<double> > *p) {
      leftPrecon = p;
    }

    /// Sets \ref rightPrecon.
    inline void setRightPrecon(Preconditioner<DOFVector<double> > *p) {
      rightPrecon = p;
    }

    /// Sets \ref marker.
    inline void setMarker(Marker *m) {
      marker = m;
    }

    /** \} */

    void writeResidualMesh(AdaptInfo *adaptInfo, const std::string name);

    /// serialization
    virtual void serialize(std::ostream &out);

    /// deserialization
    virtual void deserialize(std::istream &in);

    std::vector<FileWriterInterface*> getFileWriterList() {
      return fileWriters;
    }

  protected:
    /// Name of this problem.
    std::string name;

    /// FiniteElemSpace of this problem.
    FiniteElemSpace *feSpace_;

    /// Mesh of this problem.
    Mesh *mesh;

    /// Responsible for element marking.
    Marker *marker;

    /// Estimator of this problem. Used in \ref estimate().
    Estimator *estimator;

    /// Linear solver of this problem. Used in \ref solve().
    OEMSolver<DOFVector<double> > *solver;

    /// DOFVector storing the calculated solution of the problem.
    DOFVector<double> *solution;

    /// DOFVector for the right hand side 
    DOFVector<double> *rhs;

    /// System matrix
    DOFMatrix *systemMatrix;

    /// Matrix-vector multiplication
    MatVecMultiplier<DOFVector<double> > *matVec;

    /// Left preconditioner. Used in \ref solver.
    Preconditioner<DOFVector<double> > *leftPrecon;

    /// Right preconditioner. Used in \ref solver.
    Preconditioner<DOFVector<double> > *rightPrecon;

    /// Determines whether domain boundaries should be considered at assembling.
    bool useGetBound;

    /// Writes the meshes and solution after the adaption loop.
    std::vector<FileWriterInterface*> fileWriters;

    /** \brief
     * All actions of mesh refinement are performed by refinementManager.
     * If new refinement algorithms should be realized, one has to override
     * RefinementManager and give one instance of it to AdaptStationary.
     */
    RefinementManager *refinementManager;

    /** \brief
     * All actions of mesh coarsening are performed by coarseningManager.
     * If new coarsening algorithms should be realized, one has to override
     * CoarseningManager and give one instance of it to AdaptStationary.
     */
    CoarseningManager *coarseningManager;
  
#ifdef HAVE_PARALLEL_AMDIS
    /// Petsc application ordering to map dof indices.
    AO *applicationOrdering;

    /// Number of DOFs in the rank partition.
    int nRankDOFs;
#endif

    /// Info level.
    int info_;

  };

}

#endif
