#include "AMDiS.h"

using namespace std;
using namespace AMDiS;

// ===========================================================================
// ===== function definitions ================================================
// ===========================================================================

/** \brief
 * Dirichlet boundary function
 */
class G : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(G);

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& x) const {
    return exp(-10.0 * (x * x));
  }
};

/** \brief
 * RHS function
 */
class F : public AbstractFunction<double, WorldVector<double> >
{
public:
  MEMORY_MANAGED(F);

  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {};

  /** \brief
   * Implementation of AbstractFunction::operator().
   */
  double operator()(const WorldVector<double>& x) const {
    int dim = x.getSize();
    double r2 = (x * x);
    double ux = exp(-10.0 * r2);
    return -(400.0 * r2 - 20.0 * dim) * ux;
  };
};

// ===========================================================================
// ===== main program ========================================================
// ===========================================================================

int main(int argc, char* argv[])
{
  FUNCNAME("main");

  // ===== check for init file =====
  TEST_EXIT(argc == 2)("usage: vecellipt initfile\n");

  // ===== init parameters =====
  Parameters::init(false, argv[1]);

  // ===== create and init the scalar problem ===== 
  ProblemVec vecellipt("vecellipt");
  vecellipt.initialize(INIT_ALL);

  // === create adapt info ===
  AdaptInfo *adaptInfo = NEW AdaptInfo("vecellipt->adapt",
				       vecellipt.getNumComponents());

  // === create adapt ===
  AdaptStationary *adapt = NEW AdaptStationary("vecellipt->adapt",
					       &vecellipt,
					       adaptInfo);
  

  // ===== add boundary conditions =====
  vecellipt.addDirichletBC(1, 0, NEW G);

  // ===== create operators =====
  Operator matrixOperator00(Operator::MATRIX_OPERATOR,
			    vecellipt.getFESpace(0),
			    vecellipt.getFESpace(0));
  matrixOperator00.addSecondOrderTerm(NEW Laplace_SOT);
  vecellipt.addMatrixOperator(&matrixOperator00, 0, 0);

  Operator rhsOperator0(Operator::VECTOR_OPERATOR,
			vecellipt.getFESpace(0));

  int degree = vecellipt.getFESpace(0)->getBasisFcts()->getDegree();

  rhsOperator0.addZeroOrderTerm(NEW CoordsAtQP_ZOT(NEW F(degree)));

  vecellipt.addVectorOperator(&rhsOperator0, 0);

  Operator matrixOperator10(Operator::MATRIX_OPERATOR,
			    vecellipt.getFESpace(1),
			    vecellipt.getFESpace(0));

  Operator matrixOperator11(Operator::MATRIX_OPERATOR,
			    vecellipt.getFESpace(1),
			    vecellipt.getFESpace(1));

  matrixOperator10.addZeroOrderTerm(NEW Simple_ZOT);

  vecellipt.addMatrixOperator(&matrixOperator10, 1, 0);

  matrixOperator11.addZeroOrderTerm(NEW Simple_ZOT(-1.0));

  vecellipt.addMatrixOperator(&matrixOperator11, 1, 1);

  // ===== start adaption loop =====
  adapt->adapt();

  vecellipt.writeFiles(adaptInfo, true);

  return 0;
}


